# -*- coding: utf-8 -*-
# no-pylint: disable=missing-docstring,broad-except,too-many-instance-attributes

import json
from datetime import datetime
from os import rename
from os.path import isfile
from pgbouncemgr.logger import format_ex

STORE_ERROR = "error"
STORE_UPDATED = "updated"
STORE_NOCHANGE = "nochange"

class State():
    system_id = None
    timeline_id = None
    pgbouncer_config = None
    primary_node = None
    primary_connected = False
    primary_error = None
    err = None
    old_state = None

    def __init__(self, path):
        self.path = path
        self.load()

    def load(self):
        """Load the state from the state file.
           When this fails, an exception will be thrown."""
        # When no state file exists, we start with a clean slate.
        if not isfile(self.path):
            return

        # Otherwise, read the state from the state file.
        with open(self.path, 'r') as stream:
            try:
                state = json.load(stream)
            except json.decoder.JSONDecodeError:
                # JSON decoding failed. This is beyond repair.
                # Start with a clean slate to have the state data reinitialized.
                return

        # Copy the state over to this state object.
        self.system_id = state["system_id"]
        self.timeline_id = state["timeline_id"]
        self.pgbouncer_config = state["pgbouncer_config"]
        self.primary_node = state["primary_node"]

        # The folowing properties are always filled dynamically by the manager.
        # These are not restored from the state file on startup.
        self.primary_connected = False
        self.nodes = {}

    def is_clean_slate(self):
        return self.system_id is None

    def set_primary_connected(self, connected, err=None):
        self.primary_connected = connected
        self.primary_error = err

    def store(self, nodes):
        """Store the current state in the state file.
           Returns True when the state was stored successfully.
           Returns False when it failed. The error can be found in the err property."""
        stored_state = {
            "system_id": self.system_id,
            "timeline_id": self.timeline_id,
            "pgbouncer_config": self.pgbouncer_config,
            "primary_node": self.primary_node,
            "primary_connected": self.primary_connected,
            "primary_error": self.primary_error,
            "nodes": dict((n.name, {
                "host": n.conn_params["host"],
                "port": n.conn_params["port"],
                "status": n.status,
                "system_id": n.system_id,
                "timeline_id": n.timeline_id,
                "error": n.err}) for n in nodes)
        }

        # When the state has not changed, then don't write out a new state.
        new_state = json.dumps(stored_state, sort_keys=True, indent=2)
        if self.old_state and self.old_state == new_state:
            return STORE_NOCHANGE
        self.old_state = new_state

        try:
            self.err = None
            swap_path = "%s..SWAP" % self.path
            with open(swap_path, "w") as file_handle:
                print(new_state, file=file_handle)
            rename(swap_path, self.path)
            return STORE_UPDATED
        except Exception as exception:
            self.err = "Storing state to file (%s) failed: %s" % (
                self.path, format_exception_message(exception))
            return STORE_ERROR
